/*
   The Selection Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to use selection in OpenGL.  What I mean
   is that if the user clicks on a object with the mouse then we want the
   application to respond in some way.  In this demo when either cubes
   are clicked, then that object will start rotating.  Now to do selection
   you only need to do 4 things.  The first two is that in the render
   function you call...

   glInitName();
   glPushNames(0);

   ... before you draw anything.  Next you would call...

   glLoadName(some number);

   ... before you draw the actual object.  For example...
   
   glLoadName(1);
   auxSolidCube(2.0f);

   glLoadName(2);
   auxSolidCube(2.0f);

   The last thing you would need is a way to process the
   selection at every click of the mouse.  In this demo that is
   done in a function called ProcessObjects().
*/


// Demo: The Selection.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>        // Windows header file.
#include<gl/gl.h>          // OpenGL header.
#include<gl/glu.h>         // OpenGL utility.
#include<gl/glaux.h>       // OpenGL auxiliary.


// Function prototypes...
bool InitializeGL();
void ProcessObjects(int xPos, int yPos);
void RenderScene();
void Shutdown();


// The global device context.
HDC g_HDC;


#define CUBE1	1           // The ID for the cube.
#define CUBE2  2           // The ID for the Cube2.
#define WIDTH 640          // Window width.
#define HEIGHT 480         // Window height.


int Rotate = 0;            // 1 is to rotate just the 1st cube, 2 for the second.

// Rotation of the cube 1.
float xCubeRotation = 0.0;
float yCubeRotation = 0.0;
float zCubeRotation = 0.0;

// Rotation of the cube 2.
float xCube2Rotation = 0.0;
float yCube2Rotation = 0.0;
float zCube2Rotation = 0.0;

// Light properties for the diffuse light, specular light, and light position.
float diffuseLight[] = {0.8f, 0.8f, 0.8f, 1.0f};
float specularLight[] = {1.0f, 1.0f, 1.0f, 1.0f};
float LightPosition[] = {0.0f, 0.0f, 2.0f, 1.0f};


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.
   static POINT oldMousePos;                 // Last mouse position.
   static POINT currentMousePos;             // Current mouse position.
   static bool isMouseActive;                // Is the left mouse button down.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (float)width/(float)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;
               }
            break;
         

         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
            ProcessObjects(currentMousePos.x, currentMousePos.y);
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True if app is done, false else.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Selection Demo - By the Programming Ace.",  // window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |           // The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |               // window style.
                        WS_CLIPSIBLINGS,                             // window style.
                        100, 100,                                    // window x, y coordinate.
                        640,                                         // window width.
                        480,                                         // window height.
                        NULL,                                        // handle to parent window.
                        NULL,                                        // menu.
                        hInstance,                                   // handle to app instance.
                        NULL);                                       // ptr to window data.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   // Free all resources used by the demo.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0, 0.0, 0.0, 0.0);

   glShadeModel(GL_SMOOTH);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_LIGHTING);

   // Here we will set up a spotlight light source in our scene.
   glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
   glLightfv(GL_LIGHT0, GL_SPECULAR, specularLight);
   glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);
   glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 92.0f);
   glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, 75.0f);

   // Enable our light.
   glEnable(GL_LIGHT0);

   // Set up the material information for our objects.  Again this is just for show.
   glEnable(GL_COLOR_MATERIAL);
   glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
   glMaterialfv(GL_FRONT, GL_SPECULAR, specularLight);
   glMateriali(GL_FRONT, GL_SHININESS, 128);

   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    // Clears the screen.
	glMatrixMode(GL_MODELVIEW);                  // Set modelview matrix (save).


   // Initialize the name stack.  You will put a different number onto
   // the stack before drawing a object so if that object is clicked on
   // then you can use a switch statement to test which one it was.
   // Start off by pushing 0 on the stack.
   glInitNames();
	glPushName(0);

   // Send a number to represent the cube to glLoadName.
   glLoadName(CUBE1);

	glPushMatrix();

      // Draw the Cube.  Translate back and change the drawing color to red.
      glTranslatef(-2.0f,0.0f,-10.0f);
	   glColor3f(1.0f, 0.0f, 0.0f);

      // Rotate the object.
      glRotatef(xCubeRotation, 1.0f, 0.0f, 0.0f);
	   glRotatef(yCubeRotation, 0.0f, 1.0f, 0.0f);
	   glRotatef(zCubeRotation, 0.0f, 0.0f, 1.0f);

      // Now draw the object.
	   auxSolidCube(2.0f);

   glPopMatrix();

   // Now its time to draw the second cube.  Call glLoadName and send in the next number.
   glLoadName(CUBE2);

   glPushMatrix();
	
	   // Draw the Cube2.  Translate back and change the drawing color to blue.
	   glTranslatef(2.0f, 0.0f, -10.0f);
	   glColor3f(0.0f, 0.0f, 1.0f);

      // Rotate the object.
      glRotatef(xCube2Rotation, 1.0f, 0.0f, 0.0f);
	   glRotatef(yCube2Rotation, 0.0f, 1.0f, 0.0f);
	   glRotatef(zCube2Rotation, 0.0f, 0.0f, 1.0f);

      // Now draw the cube.
	   auxSolidCube(2.0f);

	glPopMatrix();

   // Here what I did was make it so when a object is clicked on, it will
   // start rotating.
   if(Rotate == CUBE1)
      {
         xCubeRotation += 0.2f;
         yCubeRotation += 0.5f;
         zCubeRotation += 0.5f;

         // Do error checking.
         if(xCubeRotation > 360.0f) xCubeRotation = 0.0f;
         if(yCubeRotation > 360.0f) yCubeRotation = 0.0f;
         if(zCubeRotation > 360.0f) zCubeRotation = 0.0f;
      }

   if(Rotate == CUBE2)
      {
         xCube2Rotation += 0.2f;
         yCube2Rotation += 0.5f;
         zCube2Rotation += 0.5f;

         // Do error checking.
         if(xCube2Rotation > 360.0f) xCube2Rotation = 0.0f;
         if(yCube2Rotation > 360.0f) yCube2Rotation = 0.0f;
         if(zCube2Rotation > 360.0f) zCube2Rotation = 0.0f;
      }

   SwapBuffers(g_HDC);
}


void ProcessObjects(int xPos, int yPos)
{
	GLuint select[98];   // Will hold the selection buffer.
   GLint viewport[4];   // Will hold the viewport.
	GLint clickTest;     // Will store the click test.
   int objectSelected;  // This will store the ID of the selected object.


	// Here you will set up the selection buffer.  Send in the size of the array
   // and the array itself.
	glSelectBuffer(98, select);
	
	// Place the viewport in the array viewport.
	glGetIntegerv(GL_VIEWPORT, viewport);

	glMatrixMode(GL_PROJECTION);  // Change to the projection matrix.

	glPushMatrix();

      // In order to use selection you must change the render mode to GL_SELECT.
	   glRenderMode(GL_SELECT);
	   glLoadIdentity();

      // Here we place the view volume around the mouse pointer.  This will
      // allow us to know if there is anything under the mouse.
	   gluPickMatrix(xPos, viewport[3] - yPos, 2, 2, viewport);

	   gluPerspective(45.0f, WIDTH/HEIGHT, 1.0, 100.0);

      // Draw the objects.
	   RenderScene();

      // Test if a object was clicked.
	   clickTest = glRenderMode(GL_RENDER);

	   // If a single hit occured, display the info.
	   if(clickTest == 1)
         {
	         // Get the ID of the object that was selected from the bottom of the stack.
	         objectSelected = select[3];

	         // Now depending on which was selected, it will be rotated.
	         switch(objectSelected)
		         {
		            case CUBE1:
                     // Set Rotate to 1 so that in the render the 1st cube will be rotated.
                     Rotate = 1;
			            break;

		            case CUBE2:
                     // Set Rotate to 1 so that in the render the 2nd cube will be rotated.
                     Rotate = 2;
			            break;

		            default:
			            break;
		         }
         }

	   glMatrixMode(GL_PROJECTION);  // Change to the projection matrix.

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);      // Change to modelview matrix.
}


void Shutdown()
{

}


// Recap:

// This demo doesn't show the full potential to selection.  To do selection
// this way you only need to do 4 things.  The first two is that in the render
// function you call...

// glInitName();
// glPushNames(0);

// ... before you draw anything.  Next you would call...

// glLoadName(some number);

// ... before you draw the actual object.  For example...
   
// glLoadName(1);
// auxSolidCube(2.0f);;

// glLoadName(2);
// auxSolidCube(2.0f);;

// The last thing you would need a way to process the selection at every
// click of the mouse.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com